# -*- coding: utf-8 -*-
"""DataItem represents a MICADO Data Item."""

import datetime
import os
import re
from pathlib import Path
import shutil
from typing import Tuple, List, Optional

from micado.main.commonwrapper import md5sum
from micado.main.DataPrimary import DataPrimary

# Names for layers.
KEY_SCI = "DATA"
KEY_DQ = "DQ"
KEY_ERR = "ERR"
KEY_CNT = "CNT"
KEY_PERS = "PER"
KEY_BG = "BG"
KEY_CAT = "CAT"
KEY_TAB = "TAB"
# Layer names for non-linearity.
KEY_DATAN = "DATAn"
KEY_CHI2 = "CHI2"
KEY_DOF = "DOF"


class DataItemExtra(DataPrimary):
    """DataItem represents a MICADO Data Item.

    DataItems have to represent a single Multi Extension FITS file. A DataItem
    is a DataPrimary instance composed with zero or more DataExtension
    instances.self.dpr_catg() / cls.dpr_catg_c()

    The following table describes all MICADO DataItems.
    https://docs.google.com/spreadsheets/d/1BbQkf8l8fCPSYf7skE8y0GC2UMK8kPWBkjY0ojhZqVs/edit#gid=751586500

    The columns in that table are replicated in DataItem classes as follows
    - Name: cls.eso_name()
    - Type: cls.eso_type()
    - Description: cls.description()
    - DPR CATG: self.dpr_catg() / cls.dpr_catg_c()
    - DPR TECH: self.dpr_tech() / cls.dpr_tech_c()
    - DPR TYPE: self.dpr_type() / cls.dpr_type_c()
    - DO.CATG: == DPR CATG? TODO
    - PRO CATG: == DPR CATG? TODO
    - PRO TYPE: TODO, from class hierarchy
    - #Extensions: TODO, from class hierarchy?
    - Reduction Step: TODO, from class hierarchy?
    - Internal type: TODO, is this relevant?
    """

    # creation_date: datetime.datetime = persistent(
    #     'Date this object was created [None]',
    #     datetime.datetime, datetime.datetime(1990, 1, 1))

    # extensions: List[DataExtension] = None

    def __init__(self, *args, **kwargs):
        """Initialize the DataItem."""
        DataPrimary.__init__(self, *args, **kwargs)
        self.creation_date = datetime.datetime.now()

    # TODO: Perhaps refactor these dpr_ methods?
    # E.g. with a classproperty?
    # https://stackoverflow.com/questions/5189699/how-to-make-a-class-property
    @classmethod
    def dpr_catg_c(cls):
        """DPR.CATG.

        TODO: Something sensible for non-raw data?
        """
        from micado.main.Raw import Raw

        if issubclass(cls, Raw):
            return cls.template.dpr__catg.attr.default
        else:
            return "none"

    def dpr_catg(self):
        """DPR.CATG.

        TODO: Something sensible for non-raw data?
        """
        from micado.main.Raw import Raw

        if isinstance(self, Raw):
            return self.template.dpr__catg
        else:
            return "none"

    @classmethod
    def dpr_type_c(cls):
        """DPR.TYPE."""
        from micado.main.Raw import Raw

        if issubclass(cls, Raw):
            return cls.template.dpr__type.attr.default
        else:
            return "none"

    def dpr_type(self):
        """DPR.CATG."""
        from micado.main.Raw import Raw

        if isinstance(self, Raw):
            return self.template.dpr__type
        else:
            return "none"

    @classmethod
    def dpr_tech_c(cls):
        """DPR.TECH."""
        from micado.main.Raw import Raw

        if issubclass(cls, Raw):
            return cls.template.dpr__tech.attr.default
        else:
            return "none"

    def dpr_tech(self):
        """DPR.TECH."""
        from micado.main.Raw import Raw

        if isinstance(self, Raw):
            return self.template.dpr__tech
        else:
            return "none"

    @classmethod
    def class_pro(cls):
        """The ESO PRO class, that is Raw, Img, Tab, Ext or Cat."""
        from micado.main.Raw import Raw
        from micado.main.Img import Img
        from micado.main.Tab import Tab
        from micado.main.Ext import Ext
        from micado.main.Cat import Cat

        # TODO: What should External data like persistence be called?
        #       Now it is EXT.
        for cls_pro in [Raw, Img, Tab, Ext, Cat]:
            if issubclass(cls, cls_pro):
                break
        else:
            # raise ValueError(f"{cls.__name__} is not an ESO class")
            # Return cls so we can have a name, e.g. for inheritance diagram.
            return cls

        return cls_pro

    @classmethod
    def name_file_doc(cls):
        """The name to be used as a basename for documentation.

        E.g. MasterFlat for MasterFlat.png, MasterFlat.rst, MasterFlat.tex.
        """
        # TODO: Fully settle on ESO names in the documentation, also for
        #       filenames?
        return cls.__name__

    @classmethod
    def name_oca(cls):
        """Name to use in OCA files.

        This name is generated such that it can be easily used in LIKE clauses.
        """
        # TODO: Fix hack to translate TAB into MTD.
        from micado.main.Tab import Tab

        if cls == Tab:
            return "HDR"
        base = cls.__bases__[0]
        name_base = base.name_oca() if issubclass(base, DataItemExtra) else ""
        name_self = cls.__name__.upper().replace("BASE", "")
        # OCA names are often too long for the PRO CATG keyword.
        # The value of PRO CATG can only be 54 characteres.
        # E.g. 'IMG_DEPERSISTED_DEPERSISTEDOPT_DEPERSISTEDSKY_DEPERSISTEDSIAI
        # _DEPERSISTEDSCIENCE' is too long.
        # So here are some hacks to make them shorter.
        # TODO: Move hacks elsewhere / generalize.
        name_self = name_self.replace("DEPERSISTED", "DP")
        name_self = name_self.replace("DETREND", "DT")
        name_self = name_self.replace("FINAL", "FNL")
        names = [
            n
            for n in (name_base, name_self)
            if n and "EXTRA" not in n and n not in ["DATAITEM", "FRAME"]
        ]
        the_name = "_".join(names)
        assert (
            len(the_name) <= 54
        ), f"name_oca is too long: {the_name}, {len(the_name)} > 54"
        return the_name

    @classmethod
    def name_eso(cls, includebase=False) -> str:
        """The ESO product name, e.g. MASTERDARK_IMG, DARK_RAW, ILLUM_TAB."""
        from micado.main.RecipeStepExtra import RecipeStepBase
        from micado.main.Tab import Tab

        cls_pro = cls.class_pro()
        cls_pro_name = cls_pro.__name__.upper()

        # TODO: Fix hack to translate TAB into MTD.
        if cls_pro == Tab:
            cls_pro_name = "HDR"

        if cls is cls_pro:
            # Prevent RAW_RAW.
            name = f"{cls.__name__.upper().replace('BASE', '')}"
        elif issubclass(cls, RecipeStepBase):
            # TODO: Refactor this RecipeStepBase hack.
            # Copied from RecipeStepBase
            """Name according to ESO standards. E.g. create_var_dq_layers."""
            # TODO: Refactor this duplication with Recipe.py
            # From https://stackoverflow.com/a/1176023/2097
            n1 = cls.__name__
            n2 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", n1)
            return re.sub("([a-z0-9])([A-Z])", r"\1_\2", n2).lower()
        else:
            if includebase:
                # Keep 'BASE', for inheritance diagram
                name = f"{cls.__name__.upper()}_{cls_pro_name}"
            else:
                # Remove 'BASE', only from 'PERSISTENCEBASE_IMG'.
                name = f"{cls.__name__.upper().replace('BASE', '')}_{cls_pro_name}"

        # TODO: Fix this REF hack.
        # Eg PHOTOMREF_CAT -> REF_PHOTOM_CAT
        # TODO: Propagate to name_oca()
        if "REF_" in name:
            name = "REF_" + name.replace("REF", "")

        return name

    @classmethod
    def name_eso_drl(cls) -> str:
        """Name of DRL data structure."""
        return "micado_" + cls.name_eso().lower()

    @classmethod
    def type_eso(cls):
        """The ESO type: DPR for raw data and PRO for processed data."""
        return "DPR" if cls.class_pro().name_eso() == "RAW" else "PRO"

    @classmethod
    def base_doc(cls):
        """Base class for documentation purposes.

        E.g. used for inheritance diagram.
        """
        from micado.main.DataItem import DataItem
        from micado.main.Final import Final
        from micado.main.Static import Static

        assert len(
            cls.__bases__
        ), f"Only supported for single bases {cls}, {cls.__bases__}"
        sup = cls.__bases__[0]
        if not issubclass(sup, DataItem):
            assert sup in (DataItemExtra, DataPrimary)
            return None
        elif sup.__name__.endswith("Extra") or sup in (Final, Static):
            assert len(sup.__bases__) == 1
            return sup.__bases__[0]
        else:
            return sup

    @classmethod
    def base_doc_kw(cls):
        """Get base class for documentation based on new headers."""
        base = cls.base_doc()
        if base is None:
            return None
        kws_new = base.fits_keywords_new_c()
        while kws_new == ([], []) and base is not None:
            # This base does not actually add any keywords so can be ignored.
            base = base.base_doc()
            kws_new = base.fits_keywords_new_c()

        return base

    def real_fits_keywords(self, prefixes=None):
        """Get the FITS keywords, overloaded to provide DO.CATG.

        DO.CATG is normally transient, we include it in the FITS header
        anyway.
        """
        kws = super().real_fits_keywords(prefixes=prefixes)
        if prefixes is None:
            kws["DO CATG"] = self.name_oca()
        return kws

    @classmethod
    def fits_keywords_new_c(cls):
        """The fits keywords that are new compared to base classes."""
        from micado.main.DataItem import DataItem

        sup = cls.base_doc()
        # Primary header keywords.
        kws_primary_this = cls.fits_keywords_c()
        if sup is not None:
            kws_primary_base = sup.fits_keywords_c()
            kws_primary_new = [k for k in kws_primary_this if k not in kws_primary_base]
        else:
            assert cls is DataItem
            kws_primary_new = kws_primary_this

        # Extension header keywords if available.
        if hasattr(cls, "extensions"):
            cls_ext = cls.extensions.prop_type
            kws_extension_this = cls_ext.fits_keywords_c()
            if hasattr(sup, "extensions"):
                sup_ext = sup.extensions.prop_type
                kws_extension_base = sup_ext.fits_keywords_c()
                kws_extension_new = [
                    k for k in kws_extension_this if k not in kws_extension_base
                ]
            else:
                kws_extension_new = kws_extension_this
        else:
            kws_extension_new = []
        return kws_primary_new, kws_extension_new

    @classmethod
    def layer_names(cls) -> Tuple[List[str], List[str]]:
        """The names of the layers of this data product.

        ([mandatory1, mandatory2, ..], [optional1,  optional1, ..])
        """
        # By default no name
        return [KEY_SCI], []

    @classmethod
    def layers_number(cls) -> Tuple[Optional[int], Optional[int]]:
        """Minimum/maximum number of detectors for each layer."""
        return 1, 1

    def filename_for_sof(
        self,
    ):
        """The filename to be used in SOF files"""
        from micado.main.Raw import Raw
        from micado.main.Img import Img
        from micado.main.Tab import Tab
        from micado.main.Ext import Ext
        from micado.main.Cat import Cat

        if isinstance(self, Raw):
            dt = datetime.datetime.fromtimestamp(self.mjdobs)
            snow = dt.isoformat(timespec="milliseconds")
            # Use MCD as prefix, because even MICADO makes the names too long
            # for 'ESO PRO REC1 RAW1 NAME' keywords if we want to add something
            # extra.
            sof_filename = f"MCD.{snow}.fits"
        elif isinstance(self, (Img, Tab, Ext, Cat)):
            attr_prog, cls_prog = self.__class__.get_progenitor_attr_class()
            progenitor = getattr(self, attr_prog.attribute)
            if isinstance(progenitor, list):
                progenitor = progenitor[0]

            classname = self.__class__.__name__.replace("_", "")
            # Replace "Depersisted" with "Dp" to make filenames shorter in order
            # to fit in ESO-style FITS headers.
            classname = classname.replace("Depersisted", "Dp")
            # Calibration data filenames can only be 44 characters because
            # they have to fit in the 'ESO PRO REC1 CALIB1 NAME' header.
            classname = classname.replace("Master", "Mr")

            filename_p = progenitor.filename_for_sof()
            #    MCD.1970-01-01T00:00:00.000.fits
            # or MCD.19700101000000000_DpDark.fits
            # -> MCD.19700101000000000
            assert filename_p.startswith("MCD.")
            filename_p = filename_p.split(".", maxsplit=1)[1]
            filename_p = filename_p.split("_")[0]
            filename_p = (
                filename_p.replace(".fits", "")
                .replace("-", "")
                .replace("T", "")
                .replace(":", "")
                .replace(".", "")
            )
            assert len(filename_p) == len("19700101000000000")
            sof_filename = f"MCD.{filename_p}_{classname}.fits"
        else:
            raise ValueError(f"Filename not supported for {self.__class__}.")

        return sof_filename

    def unique_filename(
        self,
        input_filename: str = None,
        extra: list = None,
    ):
        """Generate unique filename."""
        if input_filename is None:
            input_filename = self.filename

        hashsum = md5sum(input_filename)
        # 32 bit is too long usually.
        hashsum = hashsum[:8]

        sof_fn = Path(self.filename_for_sof())
        self.filename = sof_fn.with_stem(f"{sof_fn.stem}.{hashsum}").name

        if (input_filename != self.filename) and not os.path.lexists(self.filename):
            shutil.move(input_filename, self.filename)

    @classmethod
    def get_requirement(cls):
        """Get requirement classes, if any."""
        reqs = [
            getattr(cls, a).prop_type
            for a in cls.get_persistent_properties()
            if "requirement" in a
        ]
        assert len(reqs) <= 1
        return reqs[0] if reqs else None

    @classmethod
    def get_base_dataitem(cls):
        """Get the base dataitem that still uses the same recipe."""
        from micado.main.Recipe import Recipe

        if not hasattr(cls, "recipe"):
            return None
        if cls.recipe.prop_type == Recipe:
            return None
        bds = [base.get_base_dataitem() for base in cls.__bases__]
        bdsok = [
            bd
            for bd in bds
            if bd is not None and bd.recipe.prop_type == cls.recipe.prop_type
        ]
        assert len(set(bdsok)) < 2
        return bdsok[0] if bdsok else cls
