# -*- coding: utf-8 -*-
"""Meta class for all persistent classes.

The metaclass is responsible for the creation of persistent classes and their
instances.

This is derived from the common DBOBjectMeta.py to provide extra functionality.
The extra functionality might at some point be ported back to common.
"""

import itertools
import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import List, Dict, Union, Tuple, Optional
from dataclasses import dataclass, field

import yaml
from astropy.io import fits
from astropy.io.fits import Header
from micado.config.Environment import Env
from micado.main.commonwrapper import (
    link_list_property,
    desc_list_property,
    desc_property,
)

from micado.main.commonwrapper import DBSelect
from micado.main.commonwrapper import classcache
from micado.main.commonwrapper import DBObjectMeta
from micado.main.FakePropertyDict import FakePropertyDict

PATH_TEXT_YAML = Path(__file__).resolve().parent.parent.parent / "text"


@dataclass
class TextRaw:
    """Text data for Raw DataItems."""

    responsible: str = ""
    phase: str = ""
    micadosetup: str = "None"
    calunitsetup: str = "None"
    frequency: str = "None"
    duration: str = "None"
    durationper24hrs: str = "None"
    prerequisites: str = "None"
    # TODO: Find way to get processing keywords automatically.
    keywordsprocessing: List[str] = field(default_factory=lambda: [])


@dataclass
class TextPro:
    """Text data for Pro DataItems."""

    name: str = "None"
    objective: str = "None"
    frequency: str = "None"
    duration: str = ""
    procedure: str = "None"
    math: str = "None"
    remarks: str = "None"
    responsible: str = ""
    prerequisites: str = "None"
    accuracy: str = "None"
    prosepurpose: str = ""
    proseworkflow: str = ""
    # TODO: Find way to get processing keywords automatically.
    keywordsprocessing: List[str] = field(default_factory=lambda: [])


@dataclass
class TextRecipeStep:
    """Text data for RecipeSteps."""

    purpose: str = "None"
    workingremarks: str = "None"
    generaldescription: str = "None"
    mathematicaldescription: str = "None"
    qualityassessment: str = "Through QC parameters"
    errorconditions: str = "See \\cite{drlvt}."
    unittests: str = "See \\cite{drlvt}."
    inputfits: str = "None"
    inputother: str = "const hdrl_image * input\nhrdl_image * output"
    outputfits: str = "None"
    outputother: str = "cpl_error_code"


def get_meta(name):
    """Get metadata for class."""
    # TODO: Figure out better way to deal with 'depersisted'?
    fn_raw = PATH_TEXT_YAML / "raw" / f"{name.lower().replace('depersisted', '')}.yaml"
    dic_raw = yaml.safe_load(open(fn_raw).read()) if fn_raw.exists() else {}
    text_raw = TextRaw(**dic_raw)

    fn_pro = PATH_TEXT_YAML / "pro" / f"{name.lower()}.yaml"
    dic_pro = yaml.safe_load(open(fn_pro).read()) if fn_pro.exists() else {}
    text_pro = TextPro(**dic_pro)

    fn_recipestep = PATH_TEXT_YAML / "recipestep" / f"{name.lower()}.yaml"
    dic_recipestep = (
        yaml.safe_load(open(fn_recipestep).read()) if fn_recipestep.exists() else {}
    )
    text_recipestep = TextRecipeStep(**dic_recipestep)
    return text_raw, text_pro, text_recipestep


# noinspection PyAttributeOutsideInit
class MDBObjectMeta(DBObjectMeta):
    """Enhanced DBObjectMeta for MICADO.

    (M)DBObjectMeta is responsible for the creation of persistent classes
    through the __new__ method, and for the creation of persistent instances
    through the __call__ method. It would also be possible to achieve the
    same thing by overloading respectively the __init_subclass__ and __new__
    methods of DBObject. A metaclass is used because that is how it is done
    in common.

    The main enhancements above the common DBObjectMeta class are:
    - Objects can be fully defined in one expression by specifying all
      attribute values in the constructor.
    - Objects created in this way can therefore be considered immutable,
      although this is not enforced in code, only be the developers.
    - Objects are therefore also singletons: each object instantiated in this
      way will be searched for in the database instead. They will therefore
      share the same Python id and database id.
    - Objects can (and by defaul will) be committed automatically when created,
      because then they can be found again in order to ensure objects exist
      only once.

    Furthermore:
    - Objects can be instantiated directly from FITS files by converting the
      headers into attribute name-value pairs hierarchically.

    TODO: The ultimate goal is to unify this behaviour with the exists() and
    get_onthefly() methods from the OnTheFly class.
    """

    def __new__(mcs, name, bases, attributes):
        """Overloaded to set lineage and metadata."""
        cls = super().__new__(mcs, name, bases, attributes)
        cls.lineage = {}
        cls.text_raw, cls.text_pro, cls.text_recipestep = get_meta(name)
        return cls

    def __call__(
        cls,
        autocommit: bool = False,
        default_objects: bool = True,
        *args,
        **kwargs,
    ):
        """Create a MObject.

        __call__ of a metaclass performs the equivalent of __new__ and __init__
        of the class.

        __call__ dispatches based on the arguments given. The two most basic
        calls delegate to the super class DBObjectMeta:

        - DBObjectMeta.__call__(): No arguments are given.
        - DBObjectMeta.__call__(object_id): 'object_id' is given.

        The other dispatches all follow from reading a fits file, which
        recursively calls:

        - _construct_object_from_fitsfile(): filename of FITS file is given.
        - _construct_object_from_fitsheader(): FITS header is given.
        - _construct_object_from_attributes(): attributes are given as kwargs.
        - _call_other_cls(): class overide is given in CLS (for Templates).

        _construct_object_from_fitsheader() is called for each extension in the
        FITS file, as well as the primary HDU. The FITS keywords hierarchically
        describe the class structure.

        For examples, see the individual methods.

        Args:
            autocommit: The object will automatically be committed
                (if not empty).
            default_objects: Whether to use defaults that are objects.
            args: Ignored; passed on as-is to __init__ of class.
            kwargs: Information to define the desired instance. Used to
                dispatch to other _call_*() methods as described above.

        Returns:
            An MObject as defined by the header.

        TODO: Fix autocommit, see also _find_existing().
        """
        if "CLS" in kwargs:
            # The CLS is the name of a derived class. See Template
            cls_id = kwargs["CLS"]
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ["CLS"]}
            s = cls._call_other_cls(
                cls_id=cls_id,
                *args,
                **kwargs2,
            )
        elif "object_id" in kwargs:
            # A specific object id is given, delegate to super class.
            # s will be None if object_id is None, which happens when
            # instantiating an object from the database that has a
            # link_property of None.
            s = DBObjectMeta.__call__(cls, *args, **kwargs)
        elif kwargs.keys() == {"filename"}:
            filename = kwargs["filename"]
            if filename:
                s = cls._construct_object_from_fitsfile(filename=filename)
            else:
                # There are no other kwargs, but an empty filename. This is only
                # sensible for Effects that have no attributes, like
                # SummedExposure.
                # TODO: Make such classes possible.
                attr_names = [
                    a
                    for a in cls.get_persistent_properties()
                    if not a.startswith("dummy_")
                    and a not in {"filename", "globalname", "object_id"}
                ]
                # TODO: Remove backward-compatibility Recipe hack.
                from micado.main.Recipe import Recipe

                assert not attr_names or issubclass(cls, Recipe)
                s = DBObjectMeta.__call__(cls)
        elif "header" in kwargs and "CLS" in kwargs["header"]:
            # A different class is specified than this one, need to construct
            # an instance of that.
            # TODO: ESO uses ID where we use CLS, switch to that.
            header = kwargs["header"]
            cls_id = header["CLS"]
            header2 = {k: v for k, v in header.items() if k not in ["CLS"]}
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ["header"]}
            kwargs2["header"] = header2
            kwargs2["CLS"] = cls_id
            s = cls(autocommit=False, *args, **kwargs2)
        elif "header" in kwargs:
            header1 = kwargs["header"]
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ["header"]}
            s = cls._construct_object_from_fitsheader(header=header1, *args, **kwargs2)
        elif kwargs:
            s = cls._construct_object_from_attributes(
                default_objects=default_objects,
                *args,
                **kwargs,
            )
        else:
            # This is probably an attempt to construct an DBObject by
            # setting the properties manually. Therefore, do not commit.
            # Or this is called by _construct_object_from_attributes().
            s = DBObjectMeta.__call__(cls)
            autocommit = False
            from micado.main.MObjectWISE import MObject

            # Default objects should not be used when reading from FITS.
            # E.g. when no LinearityCurve is mentioned in the headers, then
            # do not use the default, instead use None.
            if not default_objects:
                for attr in s.get_persistent():
                    val = getattr(s, attr)
                    if isinstance(val, MObject):
                        setattr(s, attr, None)

        # This happens when an object with a link_property with a None value
        # is retrieved from the database.
        if s is None:
            return s

        # Commit the object if necessary.
        if not autocommit or s.persists():
            return s

        s.commit()
        # Re-retrieve the object to ensure the id is always identical.
        if s.persists():
            return cls(object_id=s.object_id.binary())
        else:
            # There probably is no database backend.
            return s

    def _construct_object_from_fitsfile(cls, filename: str = None):
        """Create an instance from file.

        Only FITS files are currently supported. The header keywords are
        converted to attributes.

        Example:
            >>> from micado.dataitems.Dark import Dark
            >>> dark = Dark(filename="mydark.fits")

        Args:
            filename: Name of the (FITS) file.

        Returns:
            An MObject instance.

        Raises:
            NotImplementedError: Reading no-extention FITS files is not yet
                supported.
        """
        # TODO: Refactor this with make_cpl()

        # A hash of the filename is linked to the filename
        # so the filename can be used in fits headers.
        if os.path.islink(filename) and not filename.endswith(".fits"):
            filename = os.readlink(filename)

        # Assume it's a fits file.
        with fits.open(filename) as hdulist:
            hduprimary = hdulist[0]
            hdus: List[fits.ImageHDU] = [
                hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)
            ]
            # "DET3.DATA" -> ('3', 'DATA')
            pattern_extname = re.compile(r"^DET(\d)\.(\w+)$")
            for hdu in hdus:
                assert "EXTNAME" in hdu.header, "Cannot find EXTNAME"
                extname = hdu.header["EXTNAME"]
                assert re.match(pattern_extname, extname), f"Bad EXTNAME {extname}"
            hdusd = {
                re.match(pattern_extname, hdu.header["EXTNAME"]).groups(): hdu
                for hdu in hdus
            }
            extensionnumbers = sorted(set(k for k, _ in hdusd))
            # E.g. ("1", "2", ..., "9")
            if hasattr(cls, "extensions"):
                prop = getattr(cls, "extensions")
                extensions = []
                for extensionnr in extensionnumbers:
                    hdudata = hdusd[(extensionnr, "DATA")]
                    extension = prop.prop_type(header=hdudata.header)
                    extensions.append(extension)

                # Committing will be dealt with by the calling function.
                s = cls(
                    autocommit=False,
                    header=hduprimary.header,
                    extensions=extensions,
                    filename=filename,
                )
            else:
                raise NotImplementedError(
                    "FITS files without extensions are not yet supported."
                )
                # TODO: support FITS files without extensions, as is required
                #       by ESO for single-layer images and for non-image
                #       data products. E.g. see
                #       dark_test.py::test_unknown_template
                #       dark_test.py::test_template
                # i.e.: s = cls(autocommit=False,
                #               header=hduprimary.header,
                #               filename=filename)

        # TODO: Fix filename is ignored stuff
        s.filename = filename
        s.load_data()
        return s

    def _construct_object_from_fitsheader(
        cls, header: Union[Header, Dict] = None, *args, **kwargs
    ):
        """Create an MObject from a FITS header.

        The FITS keywords and their values are converted into attributes.
        HIERARCH keywords are converted into MObjects recursively.

        The template are dealt with in a special matter.

        Example:
            >>> from astropy.io import fits
            >>> from micado.dataitems.DarkExtension import DarkExtension
            >>> hdu_ext = fits.open("Dark.fits")[1]
            >>> DarkExtension(header=hdu_ext.header)

        Args:
            header: FITS header, or equivalent key-value dictionary
            args: Ignored; passed on as-is.
            kwargs: Name-value pairs of persistent attributes of this class.
                Only 'filename' or 'extensions' should be in kwargs.
                See _construct_object_from_fitsfile.

        Returns:
            An MObject as defined by the header.
        """
        # kwargs might contain 'extensions', so we copy it instead of creating
        # one from scratch. See call_filename and also Raw.__init__
        kwargs2 = kwargs.copy()

        # Fix the keys.
        header0 = {_wise_key_from_fits_header_key(k): v for k, v in header.items()}
        # The headers need to be sorted to get the groupby to work.
        # The template keywords (TPL) have to go first because the template
        # is dealt with in a special matter.
        keys = sorted(sorted(header0), key=lambda k: ~k.startswith("TPL "))
        header2 = OrderedDict([(k, header0[k]) for k in keys])

        # Group the header items by the first part of the key.
        #  E.g. ig will be like
        # [
        # ...
        # ('BITPIX', [('BITPIX', -64)]),
        # ('DET',
        #  [('DET ANGLE', 0.0),
        #   ('DET DAR VALUE', 2.0),
        #   ('DET DET1 ANGLE', 0.0),
        #   ('DET DET1 GAIN', 1.0),
        #   ...]),
        #  ...,
        # ]
        ig = itertools.groupby(
            header2.items(), key=lambda ki: ki[0].split(maxsplit=1)[0]
        )
        # Convert to lists.
        header_hierarchical = OrderedDict([(k, OrderedDict(v)) for k, v in ig])

        if "PRO" in header_hierarchical:
            # Deal with the PRO RECi headers separately.
            # See ESO-VLT-DIC.PRO and pro_fits_keywords() for details.
            # PRO also contains PRO.CATG, which we ignore.
            header_pro = header_hierarchical.pop("PRO")
            # The full lineage is in the headers, we only care about the last one.
            max_rec = max(
                int(k.split()[1][3:]) for k in header_pro if k.startswith("PRO REC")
            )
            header_rec = OrderedDict(
                (k.replace(f"REC{max_rec}", "REC"), v)
                for k, v in header_pro.items()
                if k.startswith(f"PRO REC{max_rec} ")
            )

            # For real ESO data, header_rec looks like
            # OrderedDict([('PRO REC1 DRS ID', 'cpl-7.1.4'),
            #              ('PRO REC1 ID', 'micado_det_dark'),
            #              ('PRO REC1 PARAM1 NAME', 'threshup'),
            #              ('PRO REC1 PARAM1 VALUE', '0'),
            #              ('PRO REC1 PIPE ID', 'micado/0.0.1'),
            #              ('PRO REC1 RAW1 CATG', 'DEPERSISTEDDARK_IMG'),
            #              ('PRO REC1 RAW1 NAME',
            #               'MCDWS_DepersistedDark_e23d97a3ab7364c8.fits'),
            #              ('PRO REC1 RAW2 CATG', 'DEPERSISTEDDARK_IMG'),
            #              ('PRO REC1 RAW2 NAME',
            #               'MCDWS_DepersistedDark_99e37cc40d25766c.fits'),
            #              ('PRO REC1 RAW3 CATG', 'DEPERSISTEDDARK_IMG'),
            #              ('PRO REC1 RAW3 NAME',
            #               'MCDWS_DepersistedDark_ee0d8b5c8dc42fb3.fits'),
            #              ('PRO REC1 RAW4 CATG', 'DEPERSISTEDDARK_IMG'),
            #              ('PRO REC1 RAW4 NAME',
            #               'MCDWS_DepersistedDark_91ff98c8c8d26921.fits'),
            #              ('PRO REC1 RAW5 CATG', 'DEPERSISTEDDARK_IMG'),
            #              ('PRO REC1 RAW5 NAME',
            #               'MCDWS_DepersistedDark_2df073dddc973fce.fits')])

            # TODO: Do not ignore these, do something sensible.
            # TODO: Certainly use the parameters.
            keys_to_ignore = [
                "PRO REC DRS ID",
                "PRO REC ID",
                "PRO REC PARAM1 NAME",
                "PRO REC PARAM1 VALUE",
                "PRO REC PIPE ID",
            ]
            for k in keys_to_ignore:
                if k in header_rec:
                    header_rec.pop(k)

            # Group by dependency.
            # TODO: Do CALIB!
            ig = itertools.groupby(
                header_rec.items(), key=lambda ki: ki[0].rsplit(maxsplit=1)[0]
            )
            # Convert to lists.
            header_rec_d = OrderedDict(
                [(k, {k2.split()[-1]: v2 for k2, v2 in v}) for k, v in ig]
            )
            # header_rec_d is for prototype data like
            # OrderedDict([
            #     ('PRO REC1 RAW1',
            #      {'CATG': 'IMG_DP_DPDARK',
            #       'NAME': 'MCDWS_DepersistedDark_161a55d41578a6d1.fits'}),
            #     ('PRO REC1 RAW2',
            #      {'CATG': 'IMG_DP_DPDARK',
            #       'NAME': 'MCDWS_DepersistedDark_70b1cb444763784a.fits'}),
            #     ('PRO REC1 RAW3',
            #      {'CATG': 'IMG_DP_DPDARK',
            #       'NAME': 'MCDWS_DepersistedDark_eabea0f70799af0f.fits'}),
            #     ('PRO REC1 RAW4',
            #      {'CATG': 'IMG_DP_DPDARK',
            #       'NAME': 'MCDWS_DepersistedDark_4a6b4a47b07ac9e0.fits'}),
            #     ('PRO REC1 RAW5',
            #      {'CATG': 'IMG_DP_DPDARK',
            #       'NAME': 'MCDWS_DepersistedDark_9781b914a6f21e55.fits'})])
            # and for ESO files like
            # OrderedDict([
            #      ('PRO REC1 RAW1',
            #       {'CATG': 'DEPERSISTEDDARK_IMG',
            #        'NAME': 'MCDWS_DepersistedDark_e23d97a3ab7364c8.fits'}),
            #      ('PRO REC1 RAW2',
            #       {'CATG': 'DEPERSISTEDDARK_IMG',
            #        'NAME': 'MCDWS_DepersistedDark_99e37cc40d25766c.fits'}),
            #      ('PRO REC1 RAW3',
            #       {'CATG': 'DEPERSISTEDDARK_IMG',
            #        'NAME': 'MCDWS_DepersistedDark_ee0d8b5c8dc42fb3.fits'}),
            #      ('PRO REC1 RAW4',
            #       {'CATG': 'DEPERSISTEDDARK_IMG',
            #        'NAME': 'MCDWS_DepersistedDark_91ff98c8c8d26921.fits'}),
            #      ('PRO REC1 RAW5',
            #       {'CATG': 'DEPERSISTEDDARK_IMG',
            #        'NAME': 'MCDWS_DepersistedDark_2df073dddc973fce.fits'})])

            # Should be sorted:
            assert sorted(header_rec_d) == list(header_rec_d)
            attr_prog, cls_prog = cls.get_progenitor_attr_class()
            # Ensure the progenitor class is correct.
            # TODO: Only name_oca() should be used here.
            assert {d["CATG"] for k, d in header_rec_d.items() if "RAW" in k} <= {
                cls_prog.name_oca(),
                cls_prog.name_eso(),
            }

            # Convert the ESO-style headers back to a MicadoWISE-style dictionary.
            # TODO: It would be better to have one place to create this alias.
            #       For now, search for :3 to find them all.
            attr_alias = attr_prog.attribute.upper().replace("_", "")[:3]
            kwargs_progenitor = {
                # f"{attr_alias}{i} FILENAME": d['NAME']
                f"{attr_alias}{i}": OrderedDict(
                    [(f"{attr_alias}{i} FILENAME", d["NAME"])]
                )
                for i, (k, d) in enumerate(header_rec_d.items(), 1)
                if "RAW" in k
            }
            header_hierarchical.update(kwargs_progenitor)

            # TODO: Make this less fragile. It now depends on the order
            # of the CALIB keywords and it is not clear whether this is
            # guaranteed. At least the assert should fail if it doesn't work.
            kwargs_calib = {}
            attr_calibs = cls.get_calibdata()
            for i, (attr, cls_calib) in enumerate(attr_calibs.items(), 1):
                key_calib = f"PRO REC CALIB{i}"
                if key_calib not in header_rec_d:
                    continue
                rec_d = header_rec_d[key_calib]
                assert {rec_d["CATG"]} <= {cls_calib.name_oca(), cls_calib.name_eso()}
                # TODO: Search for :3 to find similar cases. Need to harmonize.
                attr_alias = attr.attribute.upper().replace("_", "")[:3]
                kwargs_calib[f"{attr_alias}"] = OrderedDict(
                    [(f"{attr_alias}{i} FILENAME", rec_d["NAME"])]
                )

            header_hierarchical.update(kwargs_calib)

            # The recipe itself should also be in the header_hierarchical,
            # but that only happens when it has process parameters. (Or well,
            # it should happen then, TODO: verify this.)
            # If there are no process parameters, then there will be nothing,
            # no the recipe will not be instantiated. Add the filename for now.
            # TODO: Find a better solution?
            if "REC" not in header_hierarchical:
                header_hierarchical["REC"] = OrderedDict([("REC FILENAME", "")])

        # TODO: Fix TEMPLATEOBHACK
        cls.temp_template = None

        # Update the kwargs
        for key, header_sub in header_hierarchical.items():
            if key == list(header_sub)[0]:
                # Normal item
                attrname, value = cls._convert_normal_keyword(key, header_sub)
                if attrname is None:
                    continue
                # INSTRUME is a FITS keyword. However, this should not
                # overwrite the instrument.
                # TODO: How to ensure this still ends up in the Template?
                from micado.main.ObservingBlock import ObservingBlock

                if attrname == "instrument" and cls not in [ObservingBlock]:
                    continue
                kwargs2[attrname] = value
            else:
                attrname, inst2, is_list = cls._convert_hierarch_keyword(
                    key, header_sub
                )
                if is_list is None:
                    continue
                # ESO headers result in a list, even if it should not.
                # E.g. HIERARCH ESO PRO REC1 RAW1 NAME even if there is
                # only one input raw.
                # It is not possible to determine whether this is the case
                # here, so that is done in _construct_object_from_attributes.
                if is_list:
                    if attrname not in kwargs2:
                        kwargs2[attrname] = []
                    kwargs2[attrname].append(inst2)
                else:
                    kwargs2[attrname] = inst2

        # Call just __call__, because one of the kwargs can be 'filename'.
        # Committing will be dealt with by the calling function.
        # default_objects=False because we do not want to use the defaults
        # for hardware that is not mentioned in the FITS headers at all.
        return cls(autocommit=False, default_objects=False, *args, **kwargs2)

    def _construct_object_from_attributes(cls, default_objects=True, *args, **kwargs):
        """Create an MObject by listing all the attribute values in the kwargs.

        The kwargs should be {attribute_name: value}, where value should either
        be a primitive, or an MObject, or input for __call__.

        Example:
            >>> from micado.hardware.DarkCurrent import DarkCurrent
            >>> dc = DarkCurrent(value=2.3)

        Args:
            default_objects: Instantiate defaults for database objects.
            args: Ignored here; passed directly to __init__ of the class.
            kwargs: Name-value pairs of persistent attributes of this class.

        Returns:
            An MObject.

        Raises:
            AttributeError: Missing or extra attributes are given.
        """
        from micado.main.Hardware import Hardware
        from micado.main.Template import Template
        from micado.main.ObservingBlock import ObservingBlock
        from micado.main.Recipe import Recipe
        from micado.main.DataItem import DataItem

        args_fits = ["image", "uncertainty", "quality", "header"]
        args_do = ["object_id", "filename", "globalname"]
        attrs_persistent = [
            a
            for a in cls.get_persistent_properties()
            # TODO: Fix filename stuff, do not inherit from DataObject?
            if not a.startswith("dummy") and a not in args_do
        ]
        props_persistent = {a: getattr(cls, a) for a in attrs_persistent}
        # Get properties that have a default value.
        classes_ok = (Hardware, Template, ObservingBlock, Recipe, DataItem)
        props_with_default = {
            a: p
            for a, p in props_persistent.items()
            if p.default is not None or issubclass(p.prop_type, classes_ok)
        }
        # The parameters that a default should be used for.
        params_default = props_with_default.keys() - kwargs.keys()

        # For now be very explicit in the necessary properties.
        # The default is used for attributes that are not specified.
        # Perhaps this can be relaxed. E.g. query for objects with any value
        # for unspecified properties.
        params_unknown = kwargs.keys() - (attrs_persistent + args_fits + args_do)
        params_missing = [
            a
            for a in attrs_persistent
            if a not in kwargs and a not in props_with_default
            # TODO: Find proper way to deal with QC parameters.
            #       Perhaps they should point to the dataitem?
            #       Idem for requirements
            and "qc" not in a and "requirement" not in a
        ]
        if params_unknown or params_missing:
            raise AttributeError(
                f"Unknown/missing attributes: {params_unknown} {params_missing}"
            )

        # Try to find an existing object.
        # This is different from other WISE-implementations.
        s_existing = cls._find_existing(**kwargs)
        if s_existing is not None:
            return s_existing

        # Create a new object. Do not commit, because properties need to be set.
        s1 = cls(autocommit=False, default_objects=default_objects)
        for name, value in kwargs.items():
            # Convert lists that should not be lists to instances. E.g. from
            # HIERARCH ESO PRO REC1 RAW1 NAME keywords
            prop = getattr(cls, name)
            from micado.main.commonwrapper import link_list_property
            from micado.main.commonwrapper import desc_list_property

            if not isinstance(
                prop, (desc_list_property, link_list_property)
            ) and isinstance(value, list):
                assert len(value) == 1
                value = value[0]

            # Mainly prevents setting None values.
            if getattr(s1, name) is not value:

                # prop is None for non-persistent attributes like 'image'
                if prop is not None:
                    # Convert value to proper type. E.g. FITS SIMPLE header
                    # is a boolean, but must be an integer in MicadoWISE.
                    ptype = (
                        prop.prop_type[0]
                        if isinstance(prop.prop_type, tuple)
                        else prop.prop_type
                    )

                    if isinstance(prop, desc_list_property):
                        value = [ptype(v) for v in value]
                    elif isinstance(prop, desc_property):
                        value = ptype(value)
                    # TODO: perhaps do something smart in an else?

                # Finally, set the value.
                setattr(s1, name, value)

        # Set defaults.
        for name in params_default:
            value = props_with_default[name].default
            from micado.main.MObjectWISE import MObject

            if not isinstance(value, MObject) or default_objects:
                setattr(s1, name, value)
            else:
                setattr(s1, name, None)

        # Allow class to override defaults above.
        s1.__init__(*args, **kwargs)

        return s1

    def _call_other_cls(cls, cls_id: str = None, *args, **kwargs):
        """Create an instance of the class specified by cls_id.

        When reading from FITS files, then the class is defined by the ID
        keyword, e.g. "TPL ID".

        Example:
            >>> from micado.main.Template import Template
            >>> from micado.templates.MicadoImgCalDark import \
                MicadoImgCalDark
            >>> header = {'DET__DIT': 1, 'DET__NDIT': 1}
            >>> template = Template(CLS="MicadoImgCalDark", header=header)
            >>> assert isinstance(template, MicadoImgCalDark)

        Args:
            cls_id: The name of the derived class.
            args: Ignored, passed on as-is.
            kwargs: Ignored, passed on as-is.

        Returns:
            An MObject instance of the specified class.
        """
        result = classcache.get(cls_id, None)
        if result is None:
            print(f"Warning: cannot find class {cls_id}")
            cls_new = cls
        else:
            cls_new = result.aclass

        # Committing will be dealt with by the calling function.
        return cls_new.__call__(autocommit=False, *args, **kwargs)

    def _find_existing(cls, **kwargs):
        """Try to find an existing object with attributes specified in kwargs.

        All attributes and their desired values should be given als kwargs
        key value pairs.

        Args:
            kwargs: name-value pairs of attributes of the object to search for.

        E.g. when cls is Detector, and kwargs is {'width': 400, 'height': 500},
        then this is equivalent to

        q = (Detector.width == 400) & (Detector.height == 500)

        TODO: Refactor this into 'exists()' and 'get_onthefly()'.

        TODO: What does it mean when attributes are not specified? That the
              value doesn't matter, or that the default should be used?

        TODO: There is currently an exception when initializing an empty
              object. However, this happens often when initializing from
              a FITS file, e.g. dark_test breaks. So at least an exception
              should be made when instantiating empty objects from FITS (and
              from the database?

        TODO: However, autocommit will automatically commit empty objects,
              which makes things even worse.

        This function is currently disabled, until some of the TODO's above
        are fixed.
        """

        return None

        if Env["mockcommon"]:
            return None

        # Select all objects and narrow the selection down one attribute at
        # a time.
        list_props_to_check = {}
        q: DBSelect = cls.select_all()
        for name, value in kwargs.items():
            prop = getattr(cls, name)
            if isinstance(prop, (link_list_property, desc_list_property)):
                list_props_to_check[prop] = value
                if value is not None and len(value):
                    for val in value:
                        q &= prop.contains(val)
            elif isinstance(prop, desc_property):
                if isinstance(prop.prop_type, tuple):
                    ptype = prop.prop_type[0]
                else:
                    ptype = prop.prop_type
                q &= prop == ptype(value)
            else:
                q &= prop == value

        # For list attributes, it is clear that all the desired values are
        # there, but not whether there are extra values or whether the order
        # matches.
        lq = list(q)
        lqok = []
        for di in lq:
            for (prop, value) in list_props_to_check.items():
                value_found = getattr(di, prop.attribute)
                if isinstance(value_found, list):
                    # Instantiate all objects in case this is a
                    # lazy_typed_list. This should not instantiate too many
                    # objects, because the "contains" queries above will have
                    # eliminated bad candidates.
                    for _ in value_found:
                        pass
                if value_found != value:
                    break
            else:
                lqok.append(di)

        # This method is always called when objects are constructed in a single
        # expression. There should therefore only be a single object that
        # has the correct set of attributes.
        lenq = len(lqok)
        if lenq > 1:
            raise ValueError(f"There should never be more than 1: {lenq}")

        toreturn = q[0] if lenq else None
        return toreturn

    def _convert_normal_keyword(
        cls, key: str, header_sub: Dict[str, Dict]
    ) -> (str, object):
        """Convert FITS header keyword and value to attribute name and value.

        header_sub should have one key and one value, e.g.
        {'BITPIX': {'BITPIX': -64}}

        The parameters to this function are similar to the parameters of the
        convert_hierarch_keyword, even though it is a bit complex for this
        function.

        Args:
            key: FITS keyword.
            header_sub: Part of FITS header converted to a dictionary.

        Returns:
            Attribute name and value.
        """
        # Import MObject locally to prevent circular imports.
        from micado.main.MObject import MObject
        from micado.main.RecipeStepExtra import name_eso_from_property

        # Get the value.
        assert len(header_sub) == 1, f"Header too long {header_sub}"
        assert len(header_sub.values()) == 1, f"Value too long {header_sub}"
        value = list(header_sub.values())[0]

        # Find the name of the corresponding property.
        attrs_existing = [
            a for a in cls.get_persistent_properties() if not a.startswith("dummy_")
        ]
        attrs_existing_max8 = {a[:8]: a for a in attrs_existing}
        props_existing = [getattr(cls, a) for a in attrs_existing]
        nameseso_existing_1 = {
            name_eso_from_property(p): p.attribute for p in props_existing
        }
        # E.g.
        # {'QC DARKMEAN': 'darkmean',
        #  'QC DARKMED': 'darkmed',
        #  'QC DARKSTD': 'darkrms',
        #  'QC NPIXSAT': 'npixsat',
        #  'QC FPIXSAT': 'fpixsat',
        #  'FILENAME': 'filename'}
        nameseso_existing_2 = {
            p.split(" ")[-1]: a for p, a in nameseso_existing_1.items()
        }
        # E.g.
        # {'DARKMEAN': 'darkmean',
        #  'DARKMED': 'darkmed',
        #  'DARKSTD': 'darkrms',
        #  'NPIXSAT': 'npixsat',
        #  'FPIXSAT': 'fpixsat',
        #  'FILENAME': 'filename'}

        if hasattr(cls, key):
            # e.g. BITPIX
            attr_name = key
        elif hasattr(cls, key.lower()):
            # e.g. filename
            attr_name = key.lower()
        elif key.lower()[:8] in attrs_existing_max8:
            # e.g. instrument
            attr_name = attrs_existing_max8[key.lower()[:8]]
        elif key in nameseso_existing_1:
            # Perhaps these should be prioritized?
            # Shouldn't normally be an issue though.
            attr_name = nameseso_existing_1[key]
        elif key in nameseso_existing_2:
            attr_name = nameseso_existing_2[key]
        else:
            print(f"Warning: {cls.__name__} has no {key}")
            return None, None

        # Convert empty cards to None.
        if isinstance(value, fits.card.Undefined):
            value = None

        # FITS headers might contain '' where None would be more appropriate.
        prop = getattr(cls, attr_name)
        if value == "" and issubclass(prop.prop_type, MObject):
            value = None

        return attr_name, value

    def _convert_hierarch_keyword(
        cls, key: str, header_sub: Dict[str, Dict]
    ) -> (str, object, bool):
        """Convert HIERARCH keyword and value to attribute name and value.

        E.g. header_sub is
        {'DET': {
            'DET ANGLE': 0.0,
            'DET DAR VALUE': 2.0,
            'DET DET1 ANGLE': 0.0,
           'DET DET1 GAIN': 1.0}

        Args:
            key: FITS keyword.
            header_sub: Part of FITS header converted to a dictionary.

        Returns:
            Attribute name and value.
        """
        # Construct the header that will be recursively passed. E.g.
        header_new = OrderedDict(
            [(k2i.split(maxsplit=1)[1], v2i) for k2i, v2i in header_sub.items()]
        )

        # Dictionary like {'DET': 'detector'}
        from micado.main.MObjectExtra import shortname_from_attrname

        attrname_from_shortname = {}
        # TODO: This part with the .upper() should be superfluous.
        for length_short in [3, 4]:
            # Most are 3 letters, some are 4, like FILT.
            attrname_from_shortname.update(
                {
                    # This part of the dictionary
                    # In general, do upper first, slice after.
                    # E.g.: "aßest"[:3].upper() == "ASSE" with 4 letters.
                    # TODO: Search for :3 to harmonize all these aliases.
                    a.upper()[:length_short].replace("_", ""): a
                    for a in cls.get_persistent_properties()
                    # Only relevant for persistent properties that are objects.
                    if isinstance(getattr(cls, a).prop_type, MDBObjectMeta)
                }
            )
        # This part with shortname_from_attrname should suffice.
        attrname_from_shortname.update(
            {
                shortname_from_attrname(a).upper(): a
                for a in cls.get_persistent_properties()
                # Only relevant for persistent properties that are objects.
                if isinstance(getattr(cls, a).prop_type, MDBObjectMeta)
            }
        )

        # The template attribute in the models is 'template', but ESO uses
        # 'TPL' as header keyword, therefore map 'TPL' to 'template' too.
        # TODO: Create better solution for TPL/TEM hack
        if "TEM" in attrname_from_shortname and "TPL" not in attrname_from_shortname:
            attrname_from_shortname["TPL"] = attrname_from_shortname["TEM"]

        # If the key is "DET1", then the attribute name is "detector" and
        # its value should be a list.
        key_text = key.strip("0123456789")

        if key in attrname_from_shortname:
            # An single object reference.
            attrname = attrname_from_shortname[key]
            prop = getattr(cls, attrname)
            # Short names should only be used for properties that refer to
            # an object themselves, never to a primitive value.
            assert isinstance(prop.prop_type, MDBObjectMeta)

            # An ObservingBlock class points to a list of Template instances.
            # However, these templates are defined at the top level of the
            # ESO FITS hierarchy. Therefore we need to artificially add the
            # Template to the ObservingBlock. The Template is already
            # constructed because the header items are ordered.
            # TODO: Fix TEMPLATEOBHACK
            if key == "OBS":
                assert cls.temp_template is not None
                value = prop.prop_type(header=header_new, templates=[cls.temp_template])
            else:
                value = prop.prop_type(header=header_new, default_objects=False)

            # If the value happens to be a Template, then cache the template
            # for reuse later.
            # TODO: Fix TEMPLATEOBHACK
            if key == "TPL":
                cls.temp_template = value

            return attrname, value, False
        elif key_text in attrname_from_shortname:
            # A list object reference.
            attrname = attrname_from_shortname[key_text]
            prop = getattr(cls, attrname)
            value = prop.prop_type(header=header_new, default_objects=False)
            return attrname, value, True
        else:
            print(f"Warning: {cls.__name__} has no {key}: {attrname_from_shortname}")
            return None, None, None

    def get_progenitor_attr_class(cls):
        """The primary progenitor of this processed data item.

        That is, the raw or processed data that is the trigger for the recipe
        that produces this dataitem to be called.
        """
        from micado.main.Raw import Raw
        from micado.main.Ext import Ext

        if issubclass(cls, (Raw, Ext)):
            return None, None

        # TODO: fix this hack
        # Many classes have a 'raw' attribute.
        for an in ["progenitor", "raw"]:
            if hasattr(cls, an):
                return getattr(cls, an), getattr(cls, an).prop_type

        if not hasattr(cls, "recipe"):
            # E.g. "Img"
            return None, None

        # TODO: Refactor this.
        # First get the input to the first RecipeStep of the Recipe
        # of this DataObject.
        progo = cls.recipe.prop_type.process.prop_type.get_first_step().get_previous()
        # Usually it is actually 'Raw', so need another way to find the real
        # raw class.
        an, prog = cls.get_raw_attr_class(cls_base=progo)
        an2 = getattr(cls, an)
        assert "Opt" not in prog.__name__
        return an2, prog

    def get_progenitor_class(cls):
        """Get just the progenitor."""
        _, prog = cls.get_progenitor_attr_class()
        return prog

    def get_raw_attr_class(
        cls, cls_base=None
    ) -> Tuple[Optional[str], Optional["MDBObjectMeta"]]:
        """Get the main raw class for a processed data item, and the attribute.

        There is no proper base class to put this in, so it is defined in the
        metaclass, even though it is not applicable to all instances of this
        metaclass.

        Args:
            cls_base: Another base class to look for than Raw.

        Returns:
            The Raw class that is the progenitor of cls.
        """
        from micado.main.Raw import Raw
        from micado.main.MObject import MObject

        cls_base = cls_base if cls_base is not None else Raw

        if issubclass(cls, cls_base):
            return None, cls

        # Do a breadth-first recursive search of the dependency graph in order
        # to find the Raw class that is the primary precursor of this processed
        # class.

        # Get all attributes.
        # TODO: walrus this
        to_search = [
            (an, getattr(cls, an).prop_type)
            for an in cls.get_persistent_properties()
            if not isinstance(getattr(cls, an).prop_type, tuple)
        ]
        # Do an order preserving sort to start with any raw classes.
        to_search.sort(key=lambda an_do: -100 if issubclass(an_do[1], cls_base) else 0)
        # Loop through all the attributes, recursively adding more, until
        # a raw class is found.
        while to_search:
            an, attrtype = to_search.pop(0)
            if isinstance(attrtype, tuple):
                # (float, int)
                continue
            if issubclass(attrtype, cls_base):
                return an, attrtype
            if issubclass(attrtype, MObject):
                to_search_more = [
                    (an, getattr(attrtype, an).prop_type)
                    for an in attrtype.get_persistent_properties()
                ]
                to_search += to_search_more
        return None, None

    def get_raw_class(cls, cls_base=None):
        """Get the main raw class for a processed data item.

        There is no proper base class to put this in, so it is defined in the
        metaclass, even though it is not applicable to all instances of this
        metaclass.

        Args:
            cls_base: Another base class to look for than Raw.

        Returns:
            The Raw class that is the progenitor of cls.
        """
        _, raw = cls.get_raw_attr_class(cls_base=cls_base)
        return raw

    def get_observingblock_name(cls):
        """Get the name of the ObservingBlock associated with this DataItem.

        Normally that is just the template name of the raw.
        """
        from micado.dataitems.StarcenterWaffle import StarcenterWaffle
        from micado.dataitems.Staroffset import Staroffset

        if issubclass(cls, (StarcenterWaffle, Staroffset)):
            return "this observing Template is part of the science Observing Block"

        raw = cls.get_raw_class()
        template = raw.template.prop_type
        from micado.main.Template import Template

        if template is Template:
            return "N/A"
        return template.name_eso()

    def get_inputdata_full(cls):
        """All the input data that."""
        # TODO: Integrate RecipeStep.get_other_inputdata_full with this.
        from micado.main.DataItem import DataItem

        return {
            getattr(cls, pp): getattr(cls, pp).prop_type
            for pp in cls.get_persistent_properties()
            if type(getattr(cls, pp).prop_type) != tuple
            and issubclass(getattr(cls, pp).prop_type, DataItem)
        }

    def get_calibdata(cls):
        """Get the input data that is not the progenitor."""
        inputdata = cls.get_inputdata_full()
        prog_attr, _ = cls.get_progenitor_attr_class()
        inputdata.pop(prog_attr)
        return inputdata

    def get_all_input_classes(cls):
        """Recursively follow inputdata and return all classes."""
        todo = [cls]
        done = []
        while todo:
            clsa = todo.pop()
            done.append(clsa)
            for _, clsb in clsa.get_inputdata_full().items():
                if clsb not in todo + done:
                    todo.append(clsb)
        return done

    def get_qc_class(cls):
        """Get the main QCParameters class for a data item.

        It is assumed that there is only one.

        Returns:
            The QCParameters class that corresponds to this data item.
        """
        from micado.dataitems.QCParameters import QCParameters

        to_search = [
            getattr(cls, an).prop_type for an in cls.get_persistent_properties()
        ]
        qcs = [
            thecls
            for thecls in to_search
            if not isinstance(thecls, tuple) and issubclass(thecls, QCParameters)
        ]
        assert len(qcs) < 2, f"To many QC classes: {qcs}"
        return qcs[0] if qcs else None

    def select_all(cls) -> DBSelect:
        """Select all instances of this class.

        This is implemented in DBObject, and thus in all instantiations of
        MDBObjectMeta. It is defined here as a forward reference.

        Returns:
            DBSelect object containing all instances.
        """
        raise NotImplementedError

    def register(cls):
        """Overloaded from DBOBjectMeta to always fill classcache.

        TODO: Implement this in DBFilebased instad.
        """
        super().register()
        if cls not in classcache:
            classcache[cls] = FakePropertyDict(cls)

    def get_derived_classes(cls, allow_extra=False):
        """Get all (imported) classes derived from this class."""
        # Must sort on something, otherwise e.g. the oca rules are in
        # a different order each time.
        clss_all = sorted(
            [classcache.get(k).aclass for k in classcache.keys()],
            key=lambda mo: mo.__name__,
        )
        return [
            c
            for c in clss_all
            if issubclass(c, cls)
            and c != cls
            and ("Extra" not in c.__name__ or allow_extra)
        ]


def _wise_key_from_fits_header_key(key: str) -> str:
    """Create a (MICADO)-WISE key from a FITS header key.

    The ESO FITS header standards prevent a perfect 1 to 1 mapping with the
    MicadoWISE data model. Fixing the key will
    - Remove leading "ESO ".
    - Prefix template specific keywords with "TPL " because the standard
      requires them to be defined at the top level of the FITS header, while
      they are actually part of the template.
    - Replace spaces with underscores in template specific keywords, because
      these are not hierarchical properties in the MicadoWISE model.

    E.g.
    - "ESO DET DIT" -> "TPL DET__DIT"

    Args:
        key: FITS header keyword.

    Returns:
        A corrected keyword
    """
    # These are keywords that are *ONLY* part of templates, but that do not
    # start with "TPL" in the ESO standard.
    keywords_template = {
        "DET DIT",
        "DET NDIT",
        "DET READOUTMODE",
        "INS MODE",
        "DPR CATG",
        "DPR TECH",
        "DPR TYPE",
    }

    if key.startswith("ESO "):
        key = key[4:]
    elif key.startswith("WISE "):
        key = key[5:]
    # TODO: Fix keywords_template hack
    if key in keywords_template:
        key = f"TPL {key.replace(' ', '__')}"
    return key
