# -*- coding: utf-8 -*-
"""Tests for Flat.py.

TODO: Add tests for different det_dit values.
"""


import shutil
import tempfile

from micado.config.Environment import Env
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.dataitems.Flat import Flat
from micado.dataitems.FlatExtension import FlatExtension
from micado.templates.MicadoImgCalFlat import MicadoImgCalFlat
from micado.test.functional.dark_test import remove_file_and_links, compare_raws
from micado.test.functional.science_test import get_filt_inst_ro
from micado.util.esokeywords import verify_fitsheaders


def generate_flat(dark: float = 2.0, size: int = 32):
    """Generate a Flat.

    Args:
        dark: Value of the dark_current.
        size: Size of the detector.
    """
    mydetector = Detector(
        x_size=size,
        y_size=size,
    )
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark)
    myshotnoise = ShotNoise(state=1)
    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        qe_curve=None,
        linearity=None,
        exposure_action=SummedExposure(),
    )

    _, _, inst, ro = get_filt_inst_ro()

    flat_template = MicadoImgCalFlat(det__dit=1, det__ndit=1)

    flat_ob = ObservingBlock(
        name="MyFlatTest",
        instrument="MICADO",
        target="flatlamp",
        templates=[flat_template],
    )

    myextensions = [
        FlatExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    flat = Flat(
        detector=mydetectorarray,
        template=flat_template,
        observing_block=flat_ob,
        instrument=inst,
        relay_optics=ro,
        extensions=myextensions,
    )
    flat.make_simulation()
    return flat


def test_flat():
    """Test creation of a Flat."""
    Env["use_python_logging"] = True
    dark = 0.0
    flat1 = generate_flat(dark=dark)

    flat1.save_fits()
    fn = flat1.filename
    verify_fitsheaders(fn)

    flat2 = Flat(filename=fn)
    compare_raws(flat1, flat2)

    fn3 = tempfile.NamedTemporaryFile(suffix=".fits", delete=False).name
    shutil.copy(fn, fn3)

    flat3 = Flat(filename=fn3)
    flat3.unique_filename()

    hash1 = flat1.filename.split("_")[-1]
    hash3 = flat3.filename.split("_")[-1]
    assert hash1 == hash3, "Bad file hashes"

    remove_file_and_links(flat1.filename)
    remove_file_and_links(flat3.filename)
    remove_file_and_links(fn3)

    # Save with self.filename set.
    flat3.save_fits()
    remove_file_and_links(flat3.filename)

    # Save while explicitly setting a filename.
    flat3.save_fits(filename=flat3.filename)
    remove_file_and_links(flat3.filename)
