# -*- coding: utf-8 -*-
"""Test IRDB compliance."""

import os
import tempfile
from pathlib import Path

import pytest
import scopesim
import yaml
from astropy.table import Table
from micado.hardware.Atmosphere import Atmosphere
from micado.hardware.Instrument import Instrument
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.Telescope import Telescope
from micado.main.HardwareThing import HardwareThing
from micado.util import irdb
from scopesim import rc
from scopesim.effects.data_container import DataContainer

from micado.hardware.DetectorArray import DetectorArray
from micado.main.Hardware import Hardware
from micado.main.MObject import MObject
from micado.util.irdb import PATH_IRDB, IRDB_FROM_MW


# TODO: Refactor this test, currently a bit messy.


def flatten_hierarchical_properties(props):
    """Flatten hierarchical property dict.

    As well as hierarchical kwargs dict, but those don't exist.

    properties :
      telescope :   ELT
      temperature : "!ATMO.temperature"
      ter_curve:
        filename : "TER_ELT_5_mirror.dat"

    to

    properties :
      telescope :   ELT
      temperature : "!ATMO.temperature"
      ter_curve__filename : "TER_ELT_5_mirror.dat"

    """
    props2 = {}
    for k, v in props.items():
        if isinstance(v, dict):
            v2 = flatten_hierarchical_properties(v)
            for kk, vv in v2.items():
                props2[f"{k}__{kk}"] = vv
        else:
            props2[k] = v
    return props2


def verify_hardware(hardware, yam):
    """Verify the attributes of hardware.

    The should be the same as the kwargs/properties in yaml,
    including default values.
    """
    assert issubclass(hardware, Hardware)
    # HardwareThings have properties and effects have kwargs.
    # They should not have both but can have neither.
    assert not ("kwargs" in yam and "properties" in yam)

    # The attributes and defaults in MicadoWISE
    mkwargs = {
        a: getattr(hardware, a).default
        for a in hardware.get_persistent_properties()
        if not a.startswith("dummy")
        and a not in ["object_id", "filename", "globalname"]
        and (
            # Annoying tuple check because prop_type can be (int, float)
            isinstance(getattr(hardware, a).prop_type, tuple)
            or not issubclass(getattr(hardware, a).prop_type, MObject)
        )
    }

    # HardwareThings have properties and effects have kwargs.
    # They should not have both but can have neither.
    assert not ("kwargs" in yam and "properties" in yam)
    key_kw = "properties" if "properties" in yam else "kwargs"

    # The attributes and defaults in the IRDB
    # NonCommonPathAberration has both pixel_scale and filename.
    # FieldConstantPSF has a warning.
    # TODO: support extra kwargs when there is also filename
    classes_to_ignore = ["NonCommonPathAberration", "FieldConstantPSF"]
    if (
        set(yam.get(key_kw, {}).keys()) == {"filename"}
        or yam.get("class", None) in classes_to_ignore
    ):
        # TODO: Fix this REALFILENAME hack
        # TODO: Check the filenames for defaults.
        if "realfilename" in mkwargs:
            assert len(mkwargs) == 1, mkwargs
            return

        dc = DataContainer(filename=yam[key_kw]["filename"])
        yam2 = {
            # Any value in the table is allowed as default.
            k: set(v.data)
            for k, v in dc.table.columns.items()
        }
    elif key_kw in yam:
        # TODO: Do something useful with filter_names?
        attrs_to_ignore = {"filter_names"}
        yam3 = flatten_hierarchical_properties(yam[key_kw])
        yam2 = {
            k: {v}
            for k, v in yam3.items()
            # Starting with an exclamation mark are defined elsewhere.
            if not (isinstance(v, str) and v.startswith("!"))
            and k not in attrs_to_ignore
        }
    else:
        # No attributes in IRDB. Just use the ones in MicadoWISE
        # TODO: do something smarter?
        yam2 = {k: {v} for k, v in mkwargs.items()}

    if "filename" in yam2:
        # Read the actual kwargs from filename.
        # However, there can be other kwargs, no idea how to deal with those.
        args_other_than_filename = yam2.keys() - set(
            [
                "filename",
            ]
        )
        args_other_than_filename_expected = {
            "active_detectors",  # from DetectorList
            "warning",  # from RelayOptics.FieldConstantPSF, but this is ignored above
        }
        assert (
            args_other_than_filename <= args_other_than_filename_expected
        ), f"{args_other_than_filename=} {args_other_than_filename_expected=}"

        assert len(yam2["filename"]) == 1, yam2["filename"]
        fn = list(yam2["filename"])[0]
        table = Table.read(PATH_IRDB / "MICADO" / fn, format="ascii")
        print(table)
        yam2 = table

    attrs_extra = mkwargs.keys() - yam2.keys()
    attrs_missing = yam2.keys() - mkwargs.keys()
    # TODO: fix these x_len and y_len
    attrs_extra_expected = {
        "name",  # Why is name not there again?
        "readout",  # Readout mode is not yet in ScopeSIM.
        # TODO: Fix realfilename hack
        # This should not occur though, because these should be caught above.
        "realfilename",
        # For Instrument apparently we added these here, but ScopeSIM does not.
        "drot",
        "plate_scale",
        "ocs__pxscale",
        # For FilterWheel
        # TODO: current_filter refers to "!OBS.filter_name_fw1", so should
        #       MicadoWISE / the Digital Design make that link too?
        "current_filter",
    }
    attrs_missing_expected = {
        # For Detector, why are these there in the first place?
        # Apparently for DetectorWindow compatibility, that we do not care about.
        "height",
        "width",
        "x",
        "y",
        # For Instrument, this should not be changed from the default.
        "filter_file_format",
        # For FilterWheel, outer unit should be implied.
        "outer_unit",
    }

    assert (
        attrs_extra - attrs_extra_expected == set()
    ), f"{attrs_extra=} {attrs_extra_expected=}"
    assert (
        attrs_missing - attrs_missing_expected == set()
    ), f"{attrs_missing=} {attrs_missing_expected=}"

    # Check default values.
    # TODO: Add test where this goes wrong? E.g. see DET REA NCHANNEL
    for k in mkwargs:
        if k not in {"id"} | attrs_extra_expected:
            assert mkwargs[k] in yam2[k], f"{k} {mkwargs[k]} {yam2[k]}"

    if issubclass(hardware, HardwareThing):
        irdb_conversion = IRDB_FROM_MW[hardware.__name__]
        attrs_effects = [
            a
            for a in hardware.get_persistent_properties()
            # Annoying tuple check because prop_type can be (int, float)
            if not isinstance(getattr(hardware, a).prop_type, tuple)
            and issubclass(getattr(hardware, a).prop_type, Hardware)
        ]
        attrs_effects_irdb = [
            irdb_conversion.get(attrname, attrname) for attrname in attrs_effects
        ]
        eff_names = [e["name"] for e in yam["effects"]]
        attrs_extra = set(attrs_effects_irdb) - set(eff_names)
        attrs_missing = set(eff_names) - set(attrs_effects_irdb)
        attrs_missing_expected = {
            # For the Detector, this should not be there in the IRDB.
            "detector_window",
            # Not relevant for MicadoWISE.
            "extra_fits_keywords",
            # For Atmosphere, we just call this tercurve so it is compatible with
            # other options.
            # TODO: Maybe harmonize.
            # "skycalc_atmosphere",
            # TODO: Properly use maps in util/irdb.py
            # "detector_linearity",
            # "full_detector_array",
            # "telescope_reflection",
        }
        attrs_extra_expected = {
            # For Atmosphere, where it is "skycalc_atmosphere" in the IRDB.
            # TODO: Ensure that this maps well.
            # "tercurve",
            # For Atmosphere, AtmosphericDispersion is commented out in the IRDB.
            # TODO: Ensure that optional (included : False in IRDB) are not used by default.
            # TODO: Or just remove them for now?
            "dispersion",
            # TODO: Properly use maps in util/irdb.py
            # "linearity",
            # "detector_array_list",
            # "reflection",
        }
        assert attrs_extra - attrs_extra_expected == set(), f"{attrs_extra=}"
        assert (
            attrs_missing - attrs_missing_expected == set()
        ), f"{attrs_missing=} {attrs_missing_expected=}"

        for (attr_e, attr_irdb) in zip(attrs_effects, attrs_effects_irdb):
            effect = getattr(hardware, attr_e).prop_type
            yam_sub = irdb.get_effect(yam, attr_irdb)
            # assert yam_sub
            verify_hardware(effect, yam_sub)


def test_irdbcompliance():
    """Test IRDB compliance."""

    verify_hardware(DetectorArray, irdb.DETECTORARRAY)
    verify_hardware(Atmosphere, irdb.ATMOSPHERE)
    verify_hardware(Telescope, irdb.TELESCOPE)
    verify_hardware(Instrument, irdb.INSTRUMENT)
    verify_hardware(RelayOptics, irdb.RELAYOPTICS)
