# -*- coding: utf-8 -*-
"""Create Python modules from a VODML/XML file.

TODO: Cleanup this script
TODO: Use type annotations
TODO: Use units
"""
import re
import textwrap
from collections import OrderedDict
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import black
from lxml import etree
from lxml.etree import Element

from scopesim.effects import DataContainer

from micado.config.Environment import Env
from micado.util.esokeywords import esokeywords
from micado.util.irdb import irdb_from_class_attr

USE_SQLALCHEMY = Env["database_engine"] == "sqlalchemy"

S_GENERATED = "This file is automatically generated."


@dataclass
class Attribute:
    """An attribute of an ObjectType."""

    name: str
    description: str = ""
    unit: str = "None"
    stype: str = ""
    default: object = ""
    vodmlid: str = ""
    esokeyword: str = ""
    is_list: bool = False
    simport: str = None

    def str_wise(self):
        """Format an attribute."""
        if isinstance(self.default, str) and self.default.endswith(")"):
            # Hack to work with class instantiations as default.
            sdefault = self.default
        elif isinstance(self.default, str) and self.default.startswith("default_"):
            sdefault = self.default.strip("'").strip('"')
        else:
            sdefault = repr(self.default).replace("'", '"')
        sattr = (
            f'    {self.name} = persistent("{self.description} '
            f'[{self.esokeyword}][{self.unit}]", {self.stype}, {sdefault})'
        )
        if len(sattr) > 120:
            ldesc = textwrap.wrap(
                f"{self.description} [{self.esokeyword}][{self.unit}]",
                width=70,
                drop_whitespace=False,
            )
            desc = "\n".join(f'        "{line}"' for line in ldesc)
            sattr = f"""    {self.name} = persistent(
{desc},
        {self.stype},
        {sdefault},
    )"""
        return sattr


@dataclass
class FitsKeyword:
    """A FITS Keyword."""

    keyword: str
    vodmlid: str
    wiseid: str


def default_objecttype():
    """Default ObjectType."""
    return [OBJECTTYPES["MObject"]]


@dataclass
class ObjectType:
    """An ObjectType."""

    name: str
    extends: "List[ObjectType]" = field(default_factory=default_objecttype)
    description: str = ""
    imports: List[str] = field(default_factory=list)
    attributes: List[Attribute] = field(default_factory=list)
    abstract: bool = False
    package: str = None
    base_and_attribute_types = None

    def issubtype(self, other):
        """Check whether this is a subtype."""
        if other.name == self.name:
            return True
        elif self.extends and any(
            isinstance(exte, ObjectType) for exte in self.extends
        ):
            return any(exte.issubtype(other) for exte in self.extends)
        else:
            return False

    def ishardware(self):
        """Is Hardware?"""
        return self.issubtype(OBJECTTYPES["Hardware"])

    def isdataitem(self):
        """Is Dataitem?"""
        return self.issubtype(OBJECTTYPES["DataItem"])

    def get_base_types(self):
        """Get base types of this type."""
        bases = []
        todo = self.extends.copy()
        while todo:
            dt = todo.pop()
            if dt not in bases:
                bases.append(dt)
                todo += dt.extends
        return bases

    def get_attribute_types(self):
        """Get types of all attributes."""
        attribute_types = []
        for attribute in self.attributes:
            if attribute.stype in ["int", "float", "str", "bool"]:
                continue
            else:
                attribute_types.append(OBJECTTYPES[attribute.stype])
        return attribute_types

    def get_base_and_attribute_types(self):
        """Recursively get all base and attribute types."""
        # Cache result.
        if self.base_and_attribute_types:
            return self.base_and_attribute_types
        the_types = [self]
        todo = self.get_base_types() + self.get_attribute_types()
        while todo:
            dt = todo.pop()
            if dt not in the_types:
                the_types.append(dt)
                todo += dt.get_base_types() + dt.get_attribute_types()
        self.base_and_attribute_types = the_types
        return the_types

    def compare_dag_order(self, other):
        """Return order in dependency graph."""
        if self == other:
            return 0
        ot1inot2 = self in other.get_base_and_attribute_types()
        ot2inot1 = other in self.get_base_and_attribute_types()
        assert not (ot1inot2 and ot2inot1), f"{self.name} {other.name}"
        if ot1inot2:
            return -1
        elif ot2inot1:
            return 1
        else:
            return 0

    def __str__(self):
        """Python version of this ObjectType."""
        if USE_SQLALCHEMY:
            return self.str_sqlalchemy()
        else:
            return self.str_wise()

    def str_sqlalchemy(self):
        """Sqlalchemy string."""
        tables = [
            f"""
association_table_{self.name.lower()}_{aa.stype.lower()} = Table(
    'association_table_{self.name.lower()}_{aa.stype.lower()}',
    Base.metadata,
    Column('{self.name.lower()}_oid', Integer, ForeignKey('{self.name.lower()}.oid')),
    Column('{aa.stype.lower()}_oid', Integer, ForeignKey('{aa.stype.lower()}.oid'))
)"""
            for aa in self.attributes
            if aa.is_list
        ]
        stables = "\n".join(tables)

        attrs = [
            f"    oid = Column(Integer, ForeignKey('{self.extends[0].name.lower()}.oid'), primary_key=True)"
        ]
        attrs += [
            f"    {aa.name}_oid = Column(Integer, ForeignKey('{aa.stype.lower()}.oid'))"
            for aa in self.attributes
            if aa.stype not in {"Integer", "Float", "String", "Boolean"}
        ]
        attrs += [
            f"    {aa.name} = Column({aa.stype})"
            if aa.stype in {"Integer", "Float", "String", "Boolean"}
            else f"    {aa.name} = relationship('{aa.stype}', "
            f"secondary='association_table_{self.name.lower()}_{aa.stype.lower()}')"
            if aa.is_list
            else f"    {aa.name} = relationship('{aa.stype}', foreign_keys=[{aa.name}_oid])"
            for aa in self.attributes
        ]

        desc_indent = "\n".join(f"    {line}" for line in self.description.split("\n"))

        simports = "\n".join(self.imports)
        sattrs = "\n\n".join(attrs)
        # Put part of the string in a separate variable so it is easier to grep
        # for generated files without finding this file.
        return "\n".join(
            [
                "# -*- coding: utf-8 -*-",
                '"""',
                f"{self.description}",
                "",
                f"{S_GENERATED} Do not modify!",
                '"""',
                "",
                f"{simports}",
                "",
                f"class {self.name}({', '.join(exte.name for exte in self.extends)}):",
                '    """',
                f"{desc_indent}",
                '    """',
                f"    __tablename__ = '{self.name.lower()}'",
                "",
                f"{sattrs}",
                "",
                """    __mapper_args__ = {{
        'polymorphic_identity': '{self.name.lower()}',
    }}""",
                "",
                stables,
                "",
            ]
        )

    def str_wise(self):
        """WISE string."""

        # In some cases, also generate a list default. It is only known at the
        # using class that a persistent property should be a list.
        # TODO: This should not take place in __str__() but earlier, where
        # the data is parsed!
        ss_default_lists = {}
        for aa in self.attributes:
            irdb_yaml = irdb_from_class_attr(self.name, aa.name)
            if irdb_yaml is not None and "kwargs" in irdb_yaml:
                if "filename" in irdb_yaml["kwargs"] and aa.default == []:
                    # So far, this only applies to
                    # DetectorArray.detector_array_list, so lets just
                    # assert that for now.
                    fn = irdb_yaml["kwargs"]["filename"]
                    dc = DataContainer(filename=fn)
                    ss_list_entries = [
                        f"    {aa.stype}({', '.join([f'{key}={row[key]}' for key in row.colnames])}),"
                        for row in dc.table
                    ]
                    newline = "\n"
                    s_default_list = f"default_{aa.name}_list = [\n{newline.join(ss_list_entries)}\n]\n\n"
                    ss_default_lists[aa.name] = s_default_list
                    aa.default = f"default_{aa.name}_list"
                    # assert aa.name != "detector_array_list"

        s_default_lists = "\n".join(ss_default_lists.values())

        # repr is necessary for defaults of string attributes
        # replace is necessary to match python_from_tsf
        attrs = [aa.str_wise() for aa in self.attributes]
        desc_lines = self.description.split("\n")
        desc_lines_short = all(len(line) < 76 for line in desc_lines)
        len_first_line = len(desc_lines[0])
        # The description should either be a single line that needs to be
        # wrapped, or it should consist of multiple lines that are already
        # wrapped correctly.
        if "\n" in self.description and not desc_lines_short:
            raise ValueError(
                f"Description lines are too long for {self.name}: {self.description}"
            )
        if len(desc_lines) > 1 and desc_lines[1]:
            raise ValueError(
                f"Second line of description for {self.name} should be empty: {self.description}"
            )

        if len(desc_lines) == 1 and len_first_line < 70:
            # Everything fits on one line with indentation and triple quotes.
            # E.g. Astrom.
            point = "" if self.description.strip().endswith(".") else "."
            desc_indent = f'''"""{self.description.strip()}{point}"""'''
        elif desc_lines_short and len_first_line < 73 and len(desc_lines) > 1:
            # The description already follows PEP-257.
            # E.g. DeriveMasterDark
            desc_indent = (
                '"""'
                + "\n".join(
                    f"    {line}".rstrip() for line in self.description.split("\n")
                ).strip()
                + '\n    """'
            )
        else:
            # It is a mess and needs to be formatted.
            # E.g. IllumCorr
            desc_wrapped = textwrap.indent(
                textwrap.fill(self.description, width=74),
                prefix="    ",
            )
            desc_indent = f'''"""{self.name}.

{desc_wrapped}
    """'''

        if self.package == "hardware":
            # Create a default instance.
            s_default_attrs = [
                f"{attr.name}={attr.default if isinstance(attr.default, str) and attr.default.startswith('default_') else repr(attr.default)},"
                for attr in self.attributes
                if not attr.name.startswith("dummy_")
            ]
            sj_default_attrs = "\n    ".join(s_default_attrs)
            s_default = f"""
default_{self.name.lower()} = {self.name}(
    {sj_default_attrs}
)
"""
        else:
            s_default = ""

        simports = "\n".join(self.imports)
        sattrs = "\n\n".join(attrs)
        # Put part of the string in a separate variable so it is easier to grep
        # for generated files without finding this file.
        return "\n".join(
            line.rstrip()
            for line in [
                "# -*- coding: utf-8 -*-",
                f'"""{self.name}.',
                "",
                f"{S_GENERATED} Do not modify!",
                '"""',
                "",
                f"{simports}",
                "",
                s_default_lists,
                "",
                "",
                f"class {self.name}({', '.join(exte.name for exte in self.extends)}):",
                f"    {desc_indent}",
                "",
                f"{sattrs}",
                "",
                f"{s_default}",
                "",
            ]
        )

    def getattr(self, attrname):
        """Get an attribute.

        TODO: actually use __getattr__? or make attributes an OrderedDict?
        """
        attrd = {a.name: a for a in self.attributes}
        return attrd.get(attrname, None)


@dataclass
class Package:
    """A Package."""

    name: str = ""
    path: Path = field(default_factory=Path)
    object_types: List[ObjectType] = field(default_factory=list)


def import_from_ref(ref, include_default=True):
    """Create import statement from reference.

    micado:hardware.DetectorArray
    from micado.hardware.DetectorArray import DetectorArray
    """
    refq, refc = ref.split(".")
    refm = ref.replace(":", ".")
    sdefaultinstance = (
        f", default_{refc.lower()}"
        if ref.startswith("micado:hardware.") and include_default
        else ""
    )

    return f"from {refm} import {refc}{sdefaultinstance}"


def parse_attribute(elattr: Element, prefix=""):
    """Parse an attribute element into an Attribute dataclass."""
    name_attr = elattr.find("name").text
    desc_attr = elattr.find("description").text
    type_attr = elattr.find("datatype").find("vodml-ref").text
    el_multi = elattr.find("multiplicity")
    if el_multi is not None:
        is_list = el_multi.find("maxOccurs").text == "-1"
    else:
        is_list = False

    if type_attr.startswith("ivoa"):
        # TODO: datetime
        type_singlet = {
            "ivoa:integer": int,
            "ivoa:real": float,
            "ivoa:IntegerQuantity": int,
            "ivoa:RealQuantity": float,
            "ivoa:string": str,
            "ivoa:boolean": bool,
        }.get(type_attr)
        type_default = {int: 0, float: 0.0, str: "", bool: False}.get(type_singlet)
        type_single = type_singlet.__name__
        if is_list:
            type_default = [type_default]
        simport = None
    else:
        type_singlet = None
        simport = import_from_ref(type_attr)
        type_single = type_attr.split(".")[-1]
        if "hardware" in type_attr:
            # type_default = f"{type_single}()"
            type_default = f"default_{type_single.lower()}"
        else:
            type_default = None
        if is_list:
            type_default = []

    vodmlid = elattr.find("vodml-id").text
    esokeyword = ""
    match = re.search("{(.*)}", desc_attr) if desc_attr else None
    # Attributes/keywords can have underscores in them. Two underscores forces
    # the use of hierarchical keywords.
    # Use first 8 characters to go from TEMPERATURE to TEMPERAT.
    name_attr_eso = name_attr.replace("__", " ").upper()[:8]
    alias_name_attr_eso = f"{prefix} {name_attr_eso}".upper()
    desc_attr_num_to_i = desc_attr.rstrip("0123456789") + "n" if desc_attr else "None"
    if desc_attr in esokeywords:
        esokeyword = desc_attr
    elif desc_attr_num_to_i in esokeywords:
        esokeyword = desc_attr_num_to_i
    elif name_attr_eso in esokeywords:
        esokeyword = name_attr_eso
    elif alias_name_attr_eso in esokeywords:
        # E.g. "HUMIDITY" of Atmosphere becomes "ATM HUMIDITY".
        esokeyword = alias_name_attr_eso

    # Values for default / description / unit.
    the_default = type_default
    desc_attr2 = desc_attr
    unit_attr = "None"

    # Override default / description / unit if the attributes is
    # an ESO keyword. (Most should be at some point.)
    if esokeyword:
        esoattr = esokeywords[esokeyword]
        the_default = esoattr.get("default", type_default)
        if the_default in (None, "None"):
            the_default = type_default
        elif type_singlet is not None:
            the_default = type_singlet(the_default)
        # Some attributes have quotes in the description.
        desc_attr2 = esoattr.get("description").replace('"', "'")
        unit_attr = esoattr.get("unit")

    # Overide the default if it is specified in curly braces.
    # These cannot be placed in the data interface dictionaries, because
    # they vary per class.
    # TODO: Do something better for these defaults.
    if match:
        default_temp = match.groups()[0]
        if type_singlet is not None:
            the_default = type_singlet(default_temp)
        else:
            the_default = default_temp

        # Update the description if it is not already set by the eso keyword.
        if not esokeyword:
            desc_attr2a = (
                desc_attr[: match.start()] + desc_attr[match.end() :]
            ).strip()
            desc_attr2 = desc_attr2a or "None"

    if desc_attr2 is not None:
        desc_attr2 = desc_attr2.strip()

    if USE_SQLALCHEMY:
        type_single = {
            "int": "Integer",
            "float": "Float",
            "str": "String",
            "bool": "Boolean",
        }.get(type_single, type_single)

    return Attribute(
        name=name_attr,
        description=desc_attr2,
        stype=type_single,
        default=the_default,
        vodmlid=vodmlid,
        unit=unit_attr,
        esokeyword=esokeyword,
        is_list=is_list,
        simport=simport,
    )


def parse_objecttype(ot: Element, package: str = None):
    """Parse one ObjectType."""
    name_ot = ot.find("name").text
    # TODO: Yet another place where the alias/short name is constructed...
    name_ot_short = name_ot[:3]

    description1 = ot.find("description").text
    # TODO: Perhaps raise if there is no description?
    description2 = description1 or "None"
    description = "\n".join(line.strip() for line in description2.split("\n"))

    element_extends = ot.find("extends")
    extends = (
        element_extends.find("vodml-ref").text
        if element_extends is not None
        else "micado:main.MObject"
    )
    extends_single = extends.split(".")[-1]
    attributes = (
        ot.findall("attribute") + ot.findall("reference") + ot.findall("composition")
    )
    is_abstract = ot.get("abstract") == "true"

    toimport = []
    if USE_SQLALCHEMY:
        toimport.append(
            import_from_ref(extends)
            + ", Table, Column, Integer, Float, String, Boolean, ForeignKey, Base, relationship"
        )
        lattrs = []
    else:
        toimport.append(
            import_from_ref(extends, include_default=False) + ", persistent"
        )
        lattrs = [
            Attribute(
                name=f"dummy_{name_ot.lower()}",
                description="dummy",
                stype="int",
                default=42,
            )
        ]

    for elattr in attributes:
        attr = parse_attribute(elattr, prefix=name_ot_short)
        if attr.simport and attr.simport not in toimport:
            toimport.append(attr.simport)
        lattrs.append(attr)

    assert extends_single in OBJECTTYPES, f"{extends_single} of {name_ot} not defined"
    assert name_ot not in OBJECTTYPES, f"{name_ot} already defined"

    otfull = ObjectType(
        name=name_ot,
        extends=[OBJECTTYPES[extends_single]],
        description=description,
        imports=toimport,
        attributes=lattrs,
        abstract=is_abstract,
        package=package,
    )
    OBJECTTYPES[otfull.name] = otfull
    print(
        f"{otfull.name} {otfull.extends[0].name if otfull.extends is not None else 'None'}"
    )
    return otfull


def parse_package(package, path_gen):
    """Parse one package."""
    name_package = package.find("name").text
    path_package = path_gen / name_package

    object_types = package.findall("objectType")

    ots = []
    for ot in object_types:
        ot = parse_objecttype(ot, package=name_package)
        ots.append(ot)

    package = Package(name=name_package, path=path_package, object_types=ots)
    return package


def main(verify=False):
    """Convert VODML/XML to Python classes."""
    p = Path(__file__).resolve().parent.parent.parent / "vodml"
    fn = p / "micado.vo-dml.xml"
    path_gen = Path(__file__).parent.parent
    assert fn.exists(), fn
    assert path_gen.exists()

    tree = etree.parse(str(fn))

    packages_xml = tree.findall("package")
    packages = [parse_package(xpackage, path_gen) for xpackage in packages_xml]
    package_names_ok = [
        "hardware",
        "dataitems",
        "recipes",
        "templates",
        "main",
        "requirements",
    ]
    packages_ok = [package for package in packages if package.name in package_names_ok]

    for package in packages_ok:
        print("PACKAGE", package.name)
        for ot in package.object_types:
            path_ot = package.path / f"{ot.name}.py"
            path_ot_extra = package.path / f"{ot.name}Extra.py"
            if path_ot_extra.is_file() and not USE_SQLALCHEMY:
                # There is an extra file that needs to be subclassed from.
                # Change
                #  class Raw(BaseFrame):
                # To
                #  class Raw(RawExtra):
                # TODO: OMG, fix this hack
                ot.extends = [ObjectType(name=ot.name + "Extra", extends=ot.extends)]
                # Change
                #  from micado.main.MObject import MObject, persistent
                # To
                #  from micado.main.MObject import persistent
                limp = ot.imports[0].split()
                ot.imports[0] = " ".join(limp[0:3] + [limp[4]])
                # Add extra import. E.g.
                #  from micado.main.RawExtra import RawExtra
                ot.imports.append(
                    f"from micado.{package.name}.{ot.name}Extra import {ot.name}Extra"
                )
            if (
                package.name == "recipes"
                and not ot.issubtype(OBJECTTYPES["RecipeStep"])
                and not ot.issubtype(OBJECTTYPES["Recipe"])
            ):
                # Hack to make RecipeSteps actual DataItems, but still
                # have them be backwards compatible with the original
                # RecipeStep class.
                # Change
                #  class CreateVarDqLayers(Img):
                # To
                #  class CreateVarDqLayers(Img, RecipeStepBase):
                # It would be better to put RecipeStepBase as the first
                # base class, because this would do the method resolution
                # in the proper order, but this is not allowed by DBMeta.py
                ot.extends.append(ObjectType(name="RecipeStepBase", extends=[]))
                ot.imports.append(
                    "from micado.main.RecipeStepExtra import RecipeStepBase"
                )
            if verify:
                with path_ot.open() as f_ot:
                    ot_expected = f_ot.read()
                    m = black.FileMode()
                    ot_expected_b = black.format_str(ot_expected, mode=m)
                    ot_generated_b = black.format_str(str(ot), mode=m)
                    # TODO: Fix RecipeStepBase hack.
                    assert (
                        ot_expected_b == ot_generated_b
                        or "RecipeStepBase" in ot_expected_b
                    )
            else:
                with path_ot.open("w", newline="\n") as f_ot:
                    f_ot.write(str(ot))

    # Sort the ObjectTypes topologically for aweimports.py.
    # This is necessary for speed when using mock common.
    ots = [ot for package in packages for ot in package.object_types]
    ots_sorted = []
    for q, ot1 in enumerate(ots):
        for i, ot2 in enumerate(ots_sorted):
            if ot2.compare_dag_order(ot1) > 0:
                ots_sorted.insert(i, ot1)
                break
        else:
            ots_sorted.append(ot1)

    s_toimport = [
        f"from micado.{ot.package}.{ot.name} import {ot.name}" for ot in ots_sorted
    ]
    s_toimport_joined = "\n".join(s_toimport)
    s_all = [
        f"    {ot.name}.__name__,"
        for package in packages
        for ot in package.object_types
    ]
    s_all_joined = "\n".join(s_all)
    s_aweimports = f'''# -*- coding: utf-8 -*-
"""Import all classes.

{S_GENERATED} Do not modify!
"""

{s_toimport_joined}

__all__ = [
{s_all_joined}
]
'''
    if not verify:
        with open(path_gen / "main" / "aweimports.py", "w") as f:
            f.write(s_aweimports)

    return packages


OBJECTTYPES = OrderedDict([("MObject", ObjectType(name="MObject", extends=[]))])

if __name__ == "__main__":
    main()
