# -*- coding: utf-8 -*-
"""Generate tsf files from Python.

TODO: Add support for hiding keywords in OHS, for DPR.TECH/TYPE/CATG e.g.
DPR.TECH.HIDE           "OHS";                     # Hide keyword in OHS
"""
import glob
import importlib
from pathlib import Path
from typing import Type

from micado.main.commonwrapper import persistent_property
from micado.main.Template import Template
from micado.toolbox.python_from_tsf import PATH_TEMPLATE_TSF, PATH_TEMPLATE_PY

from micado.util.esokeywords import esokeywords


def tsf_from_property(prop: persistent_property) -> str:
    """Create tsf parameter string from a persistent property."""
    type_from_python = {
        "double": "number",
        "integer": "integer",
        "string": "keyword",
    }

    name_eso_space_num = prop.attribute.replace("__", "_").replace("_", " ").upper()
    # Hack to get INS.FILT[12].NAME working
    name_eso_space_nonum = name_eso_space_num.replace("1", "").replace("2", "")

    # TODO: Generalize MJDOBS -> MJD-OBS hack
    if name_eso_space_nonum == "MJDOBS":
        name_eso_space_nonum = "MJD-OBS"
        name_eso_space_num = "MJD-OBS"

    # TODO: Perhaps find a more generic way to add TPL if necessary.
    if name_eso_space_nonum not in esokeywords:
        name_eso_space_nonum = f"TPL {name_eso_space_nonum}"
        name_eso_space_num = f"TPL {name_eso_space_num}"

    name_eso_dot = name_eso_space_num.replace(" ", ".")

    esokw = esokeywords[name_eso_space_nonum]
    # The DEFAULT will be what is defined as the default in the data interface
    # dictionaries. The VALUE will be the default for this particular
    # Template. E.g.
    #     INS.MODE.DEFAULT == "NODEFAULT"
    #     INS.MODE.VALUE == "IMAGING"
    svalue = prop.default
    sdefault = (
        type(svalue)(esokw["default"]) if esokw["default"] != "None" else "NODEFAULT"
    )
    assert type(svalue) == type(sdefault) or sdefault == "NODEFAULT" or svalue is None
    sparsn = [
        ("TPL.PARAM", f'"{name_eso_dot}";', "# Next template parameter"),
        (
            f"{name_eso_dot}.TYPE",
            f'"{type_from_python[esokw["type"]]}";',
            "# Keyword type",
        ),
        (f"{name_eso_dot}.RANGE", f'"{esokw["range"]}";', "# Valid range")
        if esokw["range"] != "None"
        else None,
        (f"{name_eso_dot}.DEFAULT", f'"{sdefault}";', "# Default value"),
        (f"{name_eso_dot}.VALUE", f'"{svalue}";', "# Allocated value")
        if sdefault != svalue
        else None,
        (f"{name_eso_dot}.LABEL", f'"{esokw["comment"]}";', "# Label used in P2PP"),
        (
            f"{name_eso_dot}.MINIHELP",
            f'"{esokw["description"]}";',
            "# Short Help for keyword",
        ),
    ]
    # Remove None's
    spars = [a for a in sparsn if a]
    sspars = [f"{at:23} " + f"{name:26} " + desc for at, name, desc in spars]
    return (
        """# --------------------------------------------------------------------
"""
        + "\n".join(sspars)
        + "\n"
    )


def tsf_from_template(cls: Type[Template]) -> str:
    """Create tsf file contents from Template class."""
    tpl_mode = cls.name_eso().split("_")[-1]

    header = f"""#*******************************************************************************
# E.S.O. - MICADO project
#
# Template generated automatically from {cls.name_eso()}.py using tsf_from_python.py.

PAF.HDR.START;
PAF.TYPE                "Template Signature";      # Type of PAF
PAF.ID                  "{cls.name_eso()}.tsfx";   # ID of PAF
PAF.NAME                "{cls.name_eso()}";        # Name of PAF
PAF.DESC                "{cls.__doc__.strip(".")}"; # Short description of PAF
PAF.HDR.END;

# --------------------------------------------------------------------
TPL.INSTRUM             "MCD";                     # Instrument this template is for
TPL.MODE                "{tpl_mode.upper()}";                    # Mode of observation for this template
TPL.VERSION             "1";                       # Version of the template
TPL.REFSUP              "MCD_img.ref";             # Reference Setup File
TPL.PRESEQ              "{cls.name_eso()}.seq";    # Sequencer script
TPL.GUI                 "";                        # Template GUI panel
TPL.TYPE                "{cls.dpr__catg.default.lower()}";                   # Type of template
TPL.EXECTIME            "computed";                # Expected execution time

"""

    attr_names = [
        a
        for a in cls.get_persistent_properties()
        if "dummy" not in a and a not in {"filename", "globalname", "object_id"}
    ]

    params = [tsf_from_property(getattr(cls, an)) for an in attr_names]

    footer = """
# --- oOo ---"""

    return header + "\n".join(params) + footer


def get_template_classes():
    """Import all template classes."""
    fns1 = glob.glob(str(PATH_TEMPLATE_PY / "M*.py"))
    fns1a = [fn for fn in fns1 if "Base" not in fn and "Extra" not in fn]
    fns2 = [Path(fn).parts[-1].replace(".py", "") for fn in fns1a]
    mods = {fn: importlib.import_module(f"micado.templates.{fn}") for fn in fns2}
    return [getattr(mod, fn) for fn, mod in mods.items()]


def main():
    """Generate TSF files for all Template classes."""
    classes = get_template_classes()
    for cls in classes:
        sss = tsf_from_template(cls)
        fn = PATH_TEMPLATE_TSF / f"{cls.name_eso()}.tsf"
        with open(fn, "w") as f:
            f.write(sss)


if __name__ == "__main__":
    main()
